"""
This file has generic utility methods that can be used by all parsers

These methods can rely on the os module to function, and therefore not to be used with cloud environments.

They should be used as generic utilities for any new parser that is added to the project.
"""
import os
from csv import reader
import logging

from iridauploader.parsers import exceptions
from iridauploader import model


def get_csv_reader(sample_sheet_file):

    """
    tries to create a csv.reader object which will be used to
        parse through the lines in SampleSheet.csv
    raises an error if:
            sample_sheet_file is not an existing file
            sample_sheet_file contains null byte(s)

    arguments:
            data_dir -- the directory that has SampleSheet.csv in it

    returns a csv.reader object
    """

    if os.path.isfile(sample_sheet_file):
        csv_file = open(sample_sheet_file, "r")
        # strip any trailing newline characters from the end of the line
        # including Windows newline characters (\r\n)
        csv_lines = [x.rstrip('\n') for x in csv_file]
        csv_lines = [x.rstrip('\r') for x in csv_lines]

        # open and read file in binary then send it to be parsed by csv's reader
        csv_reader = reader(csv_lines)
    else:
        raise exceptions.SampleSheetError("Sample sheet cannot be parsed as a CSV file because it's not a regular file.",
                                          sample_sheet_file)

    return csv_reader


def find_directory_list(directory):
    """Find and return all directories in the specified directory.

    Arguments:
    directory -- the directory to find directories in

    Returns: a list of directories including current directory
    """

    # Checks if we can access to the given directory, return empty and log a warning if we cannot.
    if cannot_read_directory(directory):
        raise exceptions.DirectoryError("The directory is not readable, "
                                        "can not upload samples from this directory {}".format(directory),
                                        directory)

    dir_list = next(os.walk(directory))[1]  # Gets the list of directories in the directory
    full_dir_list = []
    for d in dir_list:
        full_dir_list.append(os.path.join(directory, d))
    return full_dir_list


def build_sequencing_run_from_samples(sample_list, metadata, sequence_run_type):
    """
    Create a SequencingRun object with full project/sample/sequence_file structure

    :param sample_list: List of Sample objects
    :param metadata: metadata dict to add to the run
    :param sequence_run_type: string identifier for sequencer type when generating api route
    :return: SequencingRun
    """

    logging.debug("Building SequencingRun from parsed data")

    # create list of projects and add samples to appropriate project
    project_list = []
    for sample in sample_list:
        project = None
        for p in project_list:
            if sample.get('sample_project') == p.id:
                project = p
        if project is None:
            project = model.Project(id=sample.get('sample_project'))
            project_list.append(project)

        project.add_sample(sample)

    sequence_run = model.SequencingRun(metadata=metadata,
                                       project_list=project_list,
                                       sequencing_run_type=sequence_run_type)
    logging.debug("SequencingRun built")
    return sequence_run


def get_file_list(directory):
    """
    Get the list of file names in the data directory

    :param directory: directory to search for files
    :return: list of file names in data directory
    """
    # verify that directory exists
    if not os.path.exists(directory):
        raise exceptions.DirectoryError("Could not list files, as directory does not exist.", directory)
    # Create a file list of the directory, only hit the os once
    file_list = next(os.walk(directory))[2]
    return file_list


def cannot_read_directory(directory):
    """
    Checks if a directory is missing the filesystem Readable flag,
    :param directory:
    :return: True if unreadable, False otherwise
    """
    return not os.access(directory, os.R_OK)



class DirectoryError(Exception):
    """
    This error is thrown when the upload_status module cannot access a directory
    """

    def __init__(self, message, directory):
        """
        Initialize a IridaResourceError
        :param message: the summary message that's causing the error
        :param directory: directory that has the error
        """

        self._message = message
        self._directory = directory

    @property
    def message(self):
        return self._message

    @property
    def directory(self):
        return self._directory



import unittest
from unittest.mock import patch
from os import path
import os
import time

import iridauploader.progress as progress
from iridauploader.model import DirectoryStatus
from iridauploader.config import config

path_to_module = path.abspath(path.dirname(__file__))
if len(path_to_module) == 0:
    path_to_module = '.'


class TestGetDirectoryStatus(unittest.TestCase):
    """
    This testing class checks the different cases for new, complete, partial, and invalid sequencing run directories
    """
    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    def test_new_directory(self):
        directory = path.join(path_to_module, "new_dir")

        res = progress.get_directory_status(directory, ["SampleSheet.csv"])

        self.assertEqual(res.status, DirectoryStatus.NEW)
        self.assertEqual(res.directory, directory)
        self.assertIsNone(res.message)

    def test_new_directory_multiple_files(self):
        directory = path.join(path_to_module, "new_dir_two_files")

        res = progress.get_directory_status(directory, ["SampleSheet.csv", 'CompletedJobInfo.xml'])

        self.assertEqual(res.status, DirectoryStatus.NEW)
        self.assertEqual(res.directory, directory)
        self.assertIsNone(res.message)

    def test_new_directory_one_file_missing(self):
        directory = path.join(path_to_module, "new_dir")

        res = progress.get_directory_status(directory, ["SampleSheet.csv", 'not_a_file.txt'])

        self.assertEqual(res.status, DirectoryStatus.INVALID)
        self.assertEqual(res.directory, directory)
        self.assertIsNotNone(res.message)

    def test_new_directory_with_info_file(self):
        directory = path.join(path_to_module, "new_dir_with_info_file")

        res = progress.get_directory_status(directory, ["SampleSheet.csv"])

        self.assertEqual(res.status, DirectoryStatus.NEW)
        self.assertEqual(res.directory, directory)
        self.assertIsNone(res.message)

    def test_invalid_directory(self):
        directory = path.join(path_to_module, "invalid_dir")

        res = progress.get_directory_status(directory, ["not a SampleSheeet.csv"])

        self.assertEqual(res.status, DirectoryStatus.INVALID)
        self.assertEqual(res.directory, directory)
        self.assertIsNotNone(res.message)

    def test_inaccessible_directory(self):
        directory = path.join(path_to_module, "inaccessible_dir")

        res = progress.get_directory_status(directory, ["SampleSheet.csv"])

        self.assertEqual(res.status, DirectoryStatus.INVALID)
        self.assertEqual(res.directory, directory)
        self.assertIsNotNone(res.message)

    def test_complete_directory(self):
        directory = path.join(path_to_module, "complete_dir")

        res = progress.get_directory_status(directory, ["SampleSheet.csv"])

        self.assertEqual(res.status, DirectoryStatus.COMPLETE)
        self.assertEqual(res.directory, directory)
        self.assertIsNone(res.message)

    def test_partial_directory(self):
        directory = path.join(path_to_module, "partial_dir")

        res = progress.get_directory_status(directory, ["SampleSheet.csv"])

        self.assertEqual(res.status, DirectoryStatus.PARTIAL)
        self.assertEqual(res.directory, directory)
        self.assertIsNone(res.message)

    def test_has_miseq_complete_file(self):
        """
        Tests legacy support to show runs that contain .miseqUploaderComplete as COMPLETE
        :return:
        """
        directory = path.join(path_to_module, "has_miseq_complete")

        res = progress.get_directory_status(directory, ["SampleSheet.csv"])

        self.assertEqual(res.status, DirectoryStatus.COMPLETE)
        self.assertEqual(res.directory, directory)
        self.assertEqual(res.message, "Legacy uploader run. Set to complete to avoid uploading duplicate data.")


class TestWriteDirectoryStatus(unittest.TestCase):
    """
    This class tests that the status file is being written to correctly
    """
    directory = path.join(path_to_module, 'write_status_dir')
    status_file = path.join(directory, "irida_uploader_status.info")

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)
        config._init_config_parser()

    def tearDown(self):
        # remove status file after using it
        if path.exists(self.status_file):
            os.remove(self.status_file)

    def test_write_to_existing_file(self):
        # File does not exist yet
        self.assertFalse(path.exists(self.status_file))

        # Create DirectoryStatus to use for writing
        directory_status = DirectoryStatus(self.directory)

        # Write to file
        directory_status.status = DirectoryStatus.COMPLETE
        progress.write_directory_status(directory_status)
        # Check that file matches what we wrote
        status = progress.get_directory_status(self.directory, ["SampleSheet.csv"])
        self.assertEqual(DirectoryStatus.COMPLETE, status.status)

        # Check that the file exists now
        self.assertTrue(path.exists(self.status_file))
        # Write a new status
        directory_status.status = DirectoryStatus.ERROR
        progress.write_directory_status(directory_status)
        # Check that the file matches the status we wrote
        status = progress.get_directory_status(self.directory, ["SampleSheet.csv"])
        self.assertEqual(DirectoryStatus.ERROR, status.status)

    def test_write_to_new_file(self):
        # File does not exist yet
        self.assertFalse(path.exists(self.status_file))

        # Create DirectoryStatus to use for writing
        directory_status = DirectoryStatus(self.directory)

        # Write to file
        directory_status.status = DirectoryStatus.COMPLETE
        progress.write_directory_status(directory_status)
        # Check that file matches what we wrote
        status = progress.get_directory_status(self.directory, ["SampleSheet.csv"])
        self.assertEqual(DirectoryStatus.COMPLETE, status.status)

    def test_write_with_readonly_is_true_file(self):
        # set readonly to True in config
        config.set_config_options(readonly=True)
        # File does not exist yet
        self.assertFalse(path.exists(self.status_file))

        # Create DirectoryStatus to use for writing
        directory_status = DirectoryStatus(self.directory)

        # Try write to file (WRITE COMPLETE)
        directory_status.status = DirectoryStatus.COMPLETE
        progress.write_directory_status(directory_status)
        # Check that file matches what we wrote
        # We should find it as NEW as we have readonly=True
        status = progress.get_directory_status(self.directory, ["SampleSheet.csv"])
        self.assertEqual(DirectoryStatus.NEW, status.status)


class TestDelayedTimeHasPassed(unittest.TestCase):
    """
    Tests core.upload_helpers.delayed_time_has_passed
    """

    class StubDirectoryStatus:
        time = None

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)

    @patch("time.mktime")
    def test_has_passed(self, mock_mktime):
        """
        Tests that True is returned when enough time has passed
        :return:
        """
        # time.mktime will return the time float from 2 minutes ago
        mock_mktime.side_effect = [time.time() - (2 * 60)]

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status.time = time.strftime(DirectoryStatus.JSON_DATE_TIME_FORMAT)
        # run was delayed for 1 minute
        delay_minutes = 1

        self.assertEqual(True, progress.upload_status._delayed_time_has_passed(stub_directory_status, delay_minutes))

    def test_delay_is_zero(self):
        """
        Tests that True is returned when there is no delay
        :return:
        """
        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status.time = time.strftime(DirectoryStatus.JSON_DATE_TIME_FORMAT)
        # run was delayed for 1 minute
        delay_minutes = 0

        self.assertEqual(True, progress.upload_status._delayed_time_has_passed(stub_directory_status, delay_minutes))

    @patch("time.mktime")
    def test_has_not_passed(self, mock_mktime):
        """
        Tests that False is returned when enough time has not passed
        :return:
        """
        # time.mktime will return the time float from right now
        mock_mktime.side_effect = [time.time()]

        stub_directory_status = self.StubDirectoryStatus()
        stub_directory_status.time = time.strftime(DirectoryStatus.JSON_DATE_TIME_FORMAT)
        # run was delayed for 10 minute
        delay_minutes = 10

        self.assertEqual(False, progress.upload_status._delayed_time_has_passed(stub_directory_status, delay_minutes))



# This file contains custom widgets used by the main and config dialog windows

import logging
# PyQt needs to be imported like this because for whatever reason they decided not to include a __all__ = [...]
import PyQt5.QtWidgets as QtWidgets

import os

from iridauploader.core import logger
import iridauploader.progress as progress

from . import colours, tools


class ProgressBarHandler:
    """
    This class handles a dictionary of progress bars, indexed off of the project/sample list in the sequencing run
    It simplifies updating the progress bars percentage by allowing us to pass it only the sample and project to update
    """
    def __init__(self, q_parent=None):
        # Create a dictionary for new progress bars
        self._bar_dict = {}
        # link the qt parent so things die correctly
        self._q_parent = q_parent

    @staticmethod
    def _get_key(sample, project):
        """
        Key is created by joining the sample and project id together with a period
        :param sample:
        :param project:
        :return:
        """
        return sample + "." + project

    def _get_bar(self, sample, project):
        """
        Gets the progress bars key and returns is
        :param sample:
        :param project:
        :return:
        """
        return self._bar_dict[self._get_key(sample, project)]

    def add_bar(self, sample, project):
        """
        Create a new progress bar given a sample and project id
        :param sample: sample name
        :param project: project id
        :return: QUploadProgressBar
        """
        key = self._get_key(sample, project)
        bar = self.QUploadProgressBar(parent=self._q_parent)
        self._bar_dict[key] = bar
        return bar

    def clear(self):
        """
        Clears all the progress bars
        :return:
        """
        self._bar_dict.clear()

    def set_value(self, sample, project, value):
        """
        Sets the value of a progress bar given a sample and project id
        :param sample: sample name
        :param project: project id
        :param value: value to set (0-100), if the sample is paired end, the progress advances at half rate
        :return:
        """
        bar = self._get_bar(sample, project)
        bar.setValue(value)

    class QUploadProgressBar(QtWidgets.QProgressBar):
        # Styles for progress bar using the Qt style sheet format
        DEFAULT_STYLE = (
            "QProgressBar{{text-align: center}}"
            "QProgressBar::chunk {{background-color: {0}; width: 10px; margin: 0px;}}".format(
                colours.BLUE_LIGHT
            )
        )
        COMPLETED_STYLE = (
            "QProgressBar{{text-align: center}}"
            "QProgressBar::chunk {{background-color: {}; width: 10px; margin: 0px;}}".format(
                colours.GREEN_LIGHT
            )
        )

        def __init__(self, parent=None):
            super().__init__()
            QtWidgets.QProgressBar.__init__(self, parent)
            self.setStyleSheet(self.DEFAULT_STYLE)

        def setValue(self, value):
            """
            Sets the value of the progress bar
            :param value: int or float
            :return: None
            """
            QtWidgets.QProgressBar.setValue(self, int(value))

            # upload is complete, set the style
            if QtWidgets.QProgressBar.value(self) == 100:
                self.setStyleSheet(self.COMPLETED_STYLE)


class UploadButton(QtWidgets.QPushButton):
    def __init__(self, parent=None):
        super().__init__()
        QtWidgets.QPushButton.__init__(self, parent)
        self.set_ready()

    """
    Set various text and styles for the different states the upload button can be in
    """
    def set_ready(self):
        """
        Set the button to be blue out and ready to upload
        if there is no connection to IRIDA, show connection error instead and block
        :return:
        """
        self.setEnabled(True)
        self.setText("Start Upload")
        self.setStyleSheet("background-color: {}; color: black".format(colours.BLUE_LIGHT))

    def set_block(self):
        """
        Set the button to be greyed out and blocked
        if there is no connection to IRIDA, show connection error instead
        :return:
        """
        self.setEnabled(False)
        self.setText("Start Upload")
        self.setStyleSheet("background-color: grey; color: white")

    def set_uploading(self):
        self.setEnabled(False)
        self.setText("Uploading... See log for details.")
        self.setStyleSheet("background-color: {}; color: black".format(colours.YELLOW_LIGHT))

    def set_finished(self):
        self.setEnabled(False)
        self.setText("Upload Complete")
        self.setStyleSheet("background-color: {}; color: black".format(colours.GREEN_LIGHT))


class SampleTable(QtWidgets.QTableWidget):
    """
    This table extends the basic QTableWidget and manages the creation of Progress bars and other widgets in the table
    The widget is also responsible to handle the message passing of progress data to the Progress bar handler
    """
    # X index for the table
    TABLE_SAMPLE_NAME = 0
    TABLE_FILE_1 = 1
    TABLE_FILE_2 = 2
    TABLE_PROJECT = 3
    TABLE_PROGRESS = 4

    def __init__(self, parent=None):
        super().__init__()
        QtWidgets.QTableWidget.__init__(self, parent)
        self.setEditTriggers(QtWidgets.QTableWidget.NoEditTriggers)
        self.setColumnCount(5)
        self.setHorizontalHeaderLabels(["Sample Name", "File 1", "File 2", "Project", "Progress"])
        self.setColumnWidth(self.TABLE_SAMPLE_NAME, 170)
        self.setColumnWidth(self.TABLE_FILE_1, 200)
        self.setColumnWidth(self.TABLE_FILE_2, 200)
        self.setColumnWidth(self.TABLE_PROJECT, 70)
        self.setColumnWidth(self.TABLE_PROGRESS, 135)
        header = self.horizontalHeader()
        header.setSectionResizeMode(self.TABLE_PROGRESS, QtWidgets.QHeaderView.Stretch)
        # set up progress bar handler
        self._progress_bars = ProgressBarHandler(self)
        # subscribe the progress updates signal
        progress.signal_worker.progress_signal.connect(self._update_progress)

    def fill_table(self, sequencing_run, partial):
        """
        Given a SequencingRun, fill the table with data
        includes sample name, file names, project, and progress widget
        :param sequencing_run: SequencingRun object
        :param partial: Indicate already uploaded files in table
        :return:
        """
        # total number of samples
        sample_count = 0

        project_list = sequencing_run.project_list
        for project in project_list:
            sample_list = project.sample_list
            sample_count = sample_count + len(sample_list)

        # Set the row count to the number of samples in this run
        self.setRowCount(sample_count)
        # clear all the progress bars we may have created
        self._progress_bars.clear()

        y_index = 0
        for project in project_list:
            sample_list = project.sample_list
            for sample in sample_list:
                files = sample.sequence_file.file_list
                self.setItem(y_index, self.TABLE_SAMPLE_NAME,
                             QtWidgets.QTableWidgetItem(sample.sample_name))
                self.setItem(y_index, self.TABLE_FILE_1,
                             QtWidgets.QTableWidgetItem(os.path.basename(files[0])))
                if len(files) == 2:
                    self.setItem(y_index, self.TABLE_FILE_2,
                                 QtWidgets.QTableWidgetItem(os.path.basename(files[1])))
                self.setItem(y_index, self.TABLE_PROJECT, QtWidgets.QTableWidgetItem(project.id))

                new_progress_bar = self._progress_bars.add_bar(sample=sample.sample_name,
                                                               project=str(project.id))
                self.setCellWidget(y_index, self.TABLE_PROGRESS, new_progress_bar)
                print(sample.get_dict())
                if sample.skip and partial:
                    self._progress_bars.set_value(sample=sample.sample_name,
                                                  project=project.id,
                                                  value=100)

                y_index = y_index + 1

    def clear_table(self):
        """
        Wipes out tables contents and sets the row count back to 0
        :return:
        """
        self.clearContents()
        self.setRowCount(0)

    def _update_progress(self, data):
        """
        Updates the progress bars in the table
        receives the ProgressData object signal, and calls to the progress bar handler to update progres
        :param data: ProgressData
        :return:
        """
        self._progress_bars.set_value(sample=data.sample,
                                      project=data.project,
                                      value=data.progress)


class LogTextBox(QtWidgets.QPlainTextEdit):
    def __init__(self, parent=None):
        super().__init__()
        QtWidgets.QPlainTextEdit.__init__(self, parent)
        # Set read only mode and hide
        self.setReadOnly(True)
        self.hide()

        # Setup logging handler
        handler = tools.QtHandler()
        handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s %(message)s', datefmt='%H:%M:%S'))
        handler.setLevel(logging.INFO)
        # connect to the log/redraw function
        handler.message_writer.messageWritten.connect(self._write)
        # finally add it to the logging module
        logger.root_logger.addHandler(handler)

    def _write(self, text):
        """
        Adds the text given to the logger box, and repaints so it displays
        Used as a slot for emits
        :param text:
        :return:
        """
        self.appendPlainText(text)
        self.repaint()



import unittest
from os import path
import os

from iridauploader.tests_integration import tests_integration

import iridauploader.config as config
import iridauploader.api as api
import iridauploader.model as model
import iridauploader.progress as progress

from iridauploader.core.upload import upload_run_single_entry, batch_upload_single_entry


path_to_module = path.dirname(__file__)
if len(path_to_module) == 0:
    path_to_module = '.'


CLEANUP_DIRECTORY_LIST = [
    path.join(path_to_module, "fake_dir_data"),
    path.join(path_to_module, "fake_miniseq_data"),
    path.join(path_to_module, "fake_nextseq_data"),
    path.join(path_to_module, "fake_nextseq_nml_data"),
    path.join(path_to_module, "fake_nextseq2k_data"),
    path.join(path_to_module, "fake_ngs_data"),
    path.join(path_to_module, "fake_ngs_data_force"),
    path.join(path_to_module, "fake_ngs_data_nonexistent_project"),
    path.join(path_to_module, "fake_ngs_data_parse_fail"),
    path.join(path_to_module, "fake_ngs_data_no_completed_file"),
    path.join(path_to_module, "fake_fast5_data"),
    path.join(path_to_module, "fake_assemblies_data"),
    path.join(path_to_module, "fake_batch_data", "run_1"),
    path.join(path_to_module, "fake_batch_data", "run_2"),
    path.join(path_to_module, "fake_batch_data", "run_3")
]


class TestEndToEnd(unittest.TestCase):
    """
    This class is for End to End tests

    It uses the single entry point for the CLI to make sure all parts of the program are working together correctly
    """

    def setUp(self):
        print("\nStarting " + self.__module__ + ": " + self._testMethodName)
        # Set the config file to our test config, so that when the upload is run it grabs out config file correctly
        config.set_config_file(path.join(path_to_module, "test_config.conf"))
        config.setup()

    def tearDown(self):
        """
        Return the sample config file to blank after tests

        Deletes status file from data directories if they exist
        :return:
        """
        self.write_to_config_file("", "", "", "", "", "", False)

        for directory_path in CLEANUP_DIRECTORY_LIST:
            status_file_path = path.join(directory_path, 'irida_uploader_status.info')
            if path.exists(status_file_path):
                os.remove(status_file_path)
            log_file_path = path.join(directory_path, 'irida-uploader.log')
            if path.exists(log_file_path):
                os.remove(log_file_path)

    @staticmethod
    def write_to_config_file(client_id, client_secret, username, password, base_url, parser, readonly):
        """
        Write to out sample configuration file so that the IRIDA instance will be accessed
        :param client_id:
        :param client_secret:
        :param username:
        :param password:
        :param base_url:
        :param parser:
        :param readonly:
        :return:
        """
        config.set_config_options(client_id=client_id,
                                  client_secret=client_secret,
                                  username=username,
                                  password=password,
                                  base_url=base_url,
                                  parser=parser,
                                  readonly=readonly,
                                  delay=0,
                                  timeout=10,
                                  minimum_file_size=-1)  # allow 0kb test files to be uploaded, code blocks still run
        config.write_config_options_to_file()

    def test_valid_miseq_upload(self):
        """
        Test a valid miseq directory for upload from end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list = test_api.get_samples(project_id)

        sample_1_found = False
        sample_2_found = False
        sample_3_found = False

        for sample in sample_list:
            if sample.sample_name in ["01-1111", "02-2222", "03-3333"]:
                if sample.sample_name == "01-1111":
                    sample_1_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '01-1111_S1_L001_R1_001.fastq.gz',
                        '01-1111_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "02-2222":
                    sample_2_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '02-2222_S1_L001_R1_001.fastq.gz',
                        '02-2222_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "03-3333":
                    sample_3_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '03-3333_S1_L001_R1_001.fastq.gz',
                        '03-3333_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)
        self.assertEqual(sample_3_found, True)

    def test_batch_miseq_upload(self):
        """
        Test a valid miseq directory for upload from end to end
        We have 3 run directories
            run_1 has batch01-1111
            run_2 is invalid
            run_3 has batch03-3333
        we expect to see batch01-1111 and batch03-3333 uploaded
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_batch_project"
        project_description = "test_batch_project_description"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # Do the upload
        upload_result = batch_upload_single_entry(path.join(path_to_module, "fake_batch_data"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list = test_api.get_samples(project_id)

        sample_1_found = False
        sample_2_not_found = True
        sample_3_found = False

        for sample in sample_list:
            if sample.sample_name in ["batch01-1111", "batch02-2222", "batch03-3333"]:
                if sample.sample_name == "batch01-1111":
                    sample_1_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        'batch01-1111_S1_L001_R1_001.fastq.gz',
                        'batch01-1111_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "batch02-2222":
                    # this one should not be found
                    sample_2_not_found = False
                elif sample.sample_name == "batch03-3333":
                    sample_3_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        'batch03-3333_S1_L001_R1_001.fastq.gz',
                        'batch03-3333_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_not_found, True)
        self.assertEqual(sample_3_found, True)

    def test_read_only_miseq_upload(self):
        """
        Test upload of a readonly miseq directory with readonly turned on
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=True
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_read_only_project"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data_read_only"),
                                                force_upload=False,
                                                upload_mode=api.MODE_DEFAULT)

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list = test_api.get_samples(project_id)

        sample_1_found = False
        sample_2_found = False
        sample_3_found = False

        for sample in sample_list:
            if sample.sample_name in ["R01-1111", "R02-2222", "R03-3333"]:
                if sample.sample_name == "R01-1111":
                    sample_1_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        'R01-1111_S1_L001_R1_001.fastq.gz',
                        'R01-1111_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "R02-2222":
                    sample_2_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        'R02-2222_S1_L001_R1_001.fastq.gz',
                        'R02-2222_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "R03-3333":
                    sample_3_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        'R03-3333_S1_L001_R1_001.fastq.gz',
                        'R03-3333_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)
        self.assertEqual(sample_3_found, True)

    def test_invalid_read_only_miseq_upload(self):
        """
        Test failing to upload a readonly miseq directory because readonly is turned off
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_read_only_project_fail"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)

        # try the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data_read_only"),
                                                force_upload=False,
                                                upload_mode=api.MODE_DEFAULT)

        # Make sure the upload was a failure
        self.assertEqual(upload_result.exit_code, 1)

    def test_continue_partial_miseq_upload(self):
        """
        We set sample 2 to be already uploaded, when we verify uploads at the bottom it should be missing from the 3
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=True
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_continue_partial"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # We must set the sequencing_run_id 1 on IRIDA to uploading
        # This is because we aren't actually uploading a partially uploaded run with an existing run_id
        # If we try to upload to the already completed run_id, we will get our upload rejected
        test_api.set_seq_run_uploading(1)

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data_continue_partial"),
                                                continue_upload=True)

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list = test_api.get_samples(project_id)

        for sample in sample_list:
            if sample.sample_name in ["01thread", "02thread", "03thread"]:
                if sample.sample_name == "01thread":
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '01thread_S1_L001_R1_001.fastq.gz',
                        '01thread_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "02thread":
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    # This one is 0 because we didn't actually upload them
                    self.assertEqual(len(sequence_files), 0)
                elif sample.sample_name == "03thread":
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '03thread_S1_L001_R1_001.fastq.gz',
                        '03thread_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

    def test_valid_directory_upload(self):
        """
        Test a valid directory for upload end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="directory",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_2"
        project_description = "test_project_description_2"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_dir_data"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list = test_api.get_samples(project_id)

        sample_1_found = False
        sample_2_found = False
        sample_3_found = False

        for sample in sample_list:
            if sample.sample_name in ["my-sample-1", "my-sample-2", "my-sample-3"]:
                if sample.sample_name == "my-sample-1":
                    sample_1_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    self.assertEqual(sequence_files[0]['fileName'], 'file_1.fastq.gz')
                    self.assertEqual(sequence_files[1]['fileName'], 'file_2.fastq.gz')
                elif sample.sample_name == "my-sample-2":
                    sample_2_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    self.assertEqual(sequence_files[0]['fileName'], 'samp_F.fastq.gz')
                    self.assertEqual(sequence_files[1]['fileName'], 'samp_R.fastq.gz')
                elif sample.sample_name == "my-sample-3":
                    sample_3_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    self.assertEqual(sequence_files[0]['fileName'], 'germ_f.fastq.gz')
                    self.assertEqual(sequence_files[1]['fileName'], 'germ_r.fastq.gz')

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)
        self.assertEqual(sample_3_found, True)

    def test_valid_assemblies_directory_upload(self):
        """
        Test a valid directory of assemblies for upload end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="directory",
            readonly="False"
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_assemblies"
        project_description = "test_project_description_assemblies"
        sample_name = "my-sample-1"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # Do the upload
        upload_result = upload_run_single_entry(directory=path.join(path_to_module, "fake_assemblies_data"),
                                                force_upload=True,
                                                upload_mode=api.MODE_ASSEMBLIES)

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Get sample_id for sample we uploaded to
        sample_id = test_api.get_sample_id(sample_name, project_id)

        # Verify the files were uploaded
        sequence_files = test_api.get_assemblies_files(sample_id)
        self.assertEqual(len(sequence_files), 1)
        self.assertEqual(sequence_files[0]['fileName'], 'file_1.fasta')

    def test_valid_fast5_directory_upload(self):
        """
        Test a valid directory of fast5 for upload end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="directory",
            readonly="False"
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_fast5"
        project_description = "test_project_description_fast5"
        sample_name = "my-sample-1"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # Do the upload
        upload_result = upload_run_single_entry(directory=path.join(path_to_module, "fake_fast5_data"),
                                                force_upload=True,
                                                upload_mode=api.MODE_FAST5)

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sequence_files = test_api.get_fast5_files(project_id, sample_name)
        self.assertEqual(len(sequence_files), 1)
        self.assertEqual(sequence_files[0]['file']['fileName'], 'file_1.fast5')

    def test_valid_miniseq_upload(self):
        """
        Test a valid directory for upload end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miniseq",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_4"
        project_description = "test_project_description_4"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_miniseq_data"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list = test_api.get_samples(project_id)

        sample_1_found = False
        sample_2_found = False
        sample_3_found = False

        for sample in sample_list:
            if sample.sample_name in ["01-1111m", "02-2222m", "03-3333m"]:
                if sample.sample_name == "01-1111m":
                    sample_1_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '01-1111m_S1_L001_R1_001.fastq.gz',
                        '01-1111m_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "02-2222m":
                    sample_2_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '02-2222m_S1_L001_R1_001.fastq.gz',
                        '02-2222m_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "03-3333m":
                    sample_3_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '03-3333m_S1_L001_R1_001.fastq.gz',
                        '03-3333m_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)
        self.assertEqual(sample_3_found, True)

    def test_valid_nextseq_upload(self):
        """
        Test a valid nextseq directory for upload from end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="nextseq",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_nextseq"
        project_description = "test_project_description_nextseq"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id_1 = "1"

        # we are uploading 2 projects, so create another one
        project_name_2 = "test_project_nextseq_2"
        project_description_2 = "test_project_description_nextseq_2"
        project_2 = model.Project(name=project_name_2, description=project_description_2)
        test_api.send_project(project_2)
        project_id_2 = "2"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_nextseq_data"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list_1 = test_api.get_samples(project_id_1)
        sample_list_2 = test_api.get_samples(project_id_2)

        sample_1_found = False
        sample_2_found = False

        for sample in sample_list_1:
            if sample.sample_name == "SA20121712":
                sample_1_found = True
                sequence_files = test_api.get_sequence_files(project_id_1, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    'SA20121712_S2_R1_001.fastq.qz',
                    'SA20121712_S2_R2_001.fastq.qz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        for sample in sample_list_2:
            if sample.sample_name == "SA20121716":
                sample_2_found = True
                sequence_files = test_api.get_sequence_files(project_id_2, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    'SA20121716_S1_R1_001.fastq.qz',
                    'SA20121716_S1_R2_001.fastq.qz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)

    def test_valid_nextseq_nml_classic_upload(self):
        """
        Test a valid nextseq_nml directory for upload from end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="nextseq_nml_classic",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_nextseq_nml"
        project_description = "test_project_description_nextseq_nml"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id_1 = "1"

        # we are uploading 2 projects, so create another one
        project_name_2 = "test_project_nextseq_nml_2"
        project_description_2 = "test_project_description_nextseq_nml_2"
        project_2 = model.Project(name=project_name_2, description=project_description_2)
        test_api.send_project(project_2)
        project_id_2 = "2"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_nextseq_nml_classic_data"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list_1 = test_api.get_samples(project_id_1)
        sample_list_2 = test_api.get_samples(project_id_2)

        sample_1_found = False
        sample_2_found = False

        for sample in sample_list_1:
            if sample.sample_name == "SA20121712a":
                sample_1_found = True
                sequence_files = test_api.get_sequence_files(project_id_1, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    'SA20121712a_S2_R1_001.fastq.qz',
                    'SA20121712a_S2_R2_001.fastq.qz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        for sample in sample_list_2:
            if sample.sample_name == "SA20121716a":
                sample_2_found = True
                sequence_files = test_api.get_sequence_files(project_id_2, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    'SA20121716a_S1_R1_001.fastq.qz',
                    'SA20121716a_S1_R2_001.fastq.qz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)

    def test_valid_nextseq_nml_dual_upload(self):
        """
        Test a valid nextseq_nml directory for upload from end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="nextseq_nml_dual_upload",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_nextseq_nml_dual"
        project_description = "test_project_description_nextseq_nml_dual"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id_1 = "1"

        # we are uploading 2 projects, so create another one
        project_name_2 = "test_project_nextseq_nml_dual_2"
        project_description_2 = "test_project_description_nextseq_nml_dual_2"
        project_2 = model.Project(name=project_name_2, description=project_description_2)
        test_api.send_project(project_2)
        project_id_2 = "2"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_nextseq_nml_dual_upload_data"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list_1 = test_api.get_samples(project_id_1)
        sample_list_2 = test_api.get_samples(project_id_2)

        sample_1_found = False
        sample_2_found = False

        for sample in sample_list_1:
            if sample.sample_name == "SA20121712b":
                sample_1_found = True
                sequence_files = test_api.get_sequence_files(project_id_1, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    'SA20121712b_S2_R1_001.fastq.qz',
                    'SA20121712b_S2_R2_001.fastq.qz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        for sample in sample_list_2:
            if sample.sample_name == "SA20121716b":
                sample_2_found = True
                sequence_files = test_api.get_sequence_files(project_id_2, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    'SA20121716b_S1_R1_001.fastq.qz',
                    'SA20121716b_S1_R2_001.fastq.qz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)

    def test_valid_nextseq_nml_strict_sample_name_matching(self):
        """
        Test a valid nextseq_nml directory for upload from end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="nextseq_nml_strict_sample_name",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_nextseq_nml_strict"
        project_description = "test_project_description_nextseq_nml_strict"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id_1 = "1"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_nextseq_data_nml_strict_sample_name"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list_1 = test_api.get_samples(project_id_1)

        sample_1_found = False
        sample_2_found = False

        for sample in sample_list_1:
            if sample.sample_name == "SA20121712c":
                sample_1_found = True
                sequence_files = test_api.get_sequence_files(project_id_1, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    'SA20121712c_S2_R1_001.fastq.qz',
                    'SA20121712c_S2_R2_001.fastq.qz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
            if sample.sample_name == "other-SA20121712c":
                sample_2_found = True
                sequence_files = test_api.get_sequence_files(project_id_1, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    'other-SA20121712c_S1_R1_001.fastq.qz',
                    'other-SA20121712c_S1_R2_001.fastq.qz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)

    def test_valid_nextseq2k_upload(self):
        """
        Test a valid nextseq2k directory for upload from end to end
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="nextseq2k_nml",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_nextseq2k"
        project_description = "test_project_description_nextseq2k"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id_1 = "1"

        # we are uploading 2 projects, so create another one
        project_name_2 = "test_project_nextseq2k_2"
        project_description_2 = "test_project_description_nextseq2k_2"
        project_2 = model.Project(name=project_name_2, description=project_description_2)
        test_api.send_project(project_2)
        project_id_2 = "2"

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_nextseq2k_data"))

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list_1 = test_api.get_samples(project_id_1)
        sample_list_2 = test_api.get_samples(project_id_2)

        sample_1_found = False
        sample_2_found = False

        for sample in sample_list_1:
            if sample.sample_name == "01A100001":
                sample_1_found = True
                sequence_files = test_api.get_sequence_files(project_id_1, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    '01A100001_S1_L001_R1_001.fastq.gz',
                    '01A100001_S1_L001_R2_001.fastq.gz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        for sample in sample_list_2:
            if sample.sample_name == "01A100002":
                sample_2_found = True
                sequence_files = test_api.get_sequence_files(project_id_2, sample.sample_name)
                self.assertEqual(len(sequence_files), 2)
                res_sequence_file_names = [
                    sequence_files[0]['fileName'],
                    sequence_files[1]['fileName']
                ]
                expected_sequence_file_names = [
                    '01A100002_S1_L001_R1_001.fastq.gz',
                    '01A100002_S1_L001_R2_001.fastq.gz'
                ]
                self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)

    def test_upload_to_nonexistent_project(self):
        """
        Everything is correct except the sample sheet file specifies an invalid project
        Samples should not be uploaded
        :return:
        """
        # try to upload to a non existent project
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data_nonexistent_project"))

        # Make sure the upload was a failure
        self.assertEqual(upload_result.exit_code, 1)

        # Verify that the project does not exist
        project_id = "1000"
        with self.assertRaises(api.exceptions.IridaResourceError):
            test_api.get_samples(project_id)

    def test_upload_parse_fail(self):
        """
        Given an invalid sample sheet, make sure that the upload does not happen
        :return:
        """
        # try to upload to a non existent project
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)

        # Do the upload
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data_parse_fail"))

        # Make sure the upload was a failure
        self.assertEqual(upload_result.exit_code, 1)

    def test_valid_miseq_with_status_file_force(self):
        """
        Test a valid miseq directory for upload from end to end
        We create a status file that indicates the files have already been uploaded,
        and then use the force option to upload anyways
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=False
        )

        # instance an api
        test_api = api.ApiCalls(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            base_url=tests_integration.base_url,
            username=tests_integration.username,
            password=tests_integration.password
        )

        # Create a test project, the uploader does not make new projects on its own
        # so one must exist to upload samples into
        # This may not be the project that the files get uploaded to,
        # but one will be made in the case this is the only test being run
        project_name = "test_project_3"
        project_description = "test_project_description"
        project = model.Project(name=project_name, description=project_description)
        test_api.send_project(project)
        # We always upload to project "1" so that tests will be consistent no matter how many / which tests are run
        project_id = "1"

        # Write a status file to the upload directory that we can force past
        directory_status = model.DirectoryStatus(directory=path.join(path_to_module, "fake_ngs_data_force"))
        directory_status.status = model.DirectoryStatus.COMPLETE
        progress.write_directory_status(directory_status)

        # Do the upload, with force option
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data_force"), True)

        # Make sure the upload was a success
        self.assertEqual(upload_result.exit_code, 0)

        # Verify the files were uploaded
        sample_list = test_api.get_samples(project_id)

        sample_1_found = False
        sample_2_found = False
        sample_3_found = False

        for sample in sample_list:
            if sample.sample_name in ["01-1111", "02-2222", "03-3333"]:
                if sample.sample_name == "01-1111":
                    sample_1_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '01-1111_S1_L001_R1_001.fastq.gz',
                        '01-1111_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "02-2222":
                    sample_2_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '02-2222_S1_L001_R1_001.fastq.gz',
                        '02-2222_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())
                elif sample.sample_name == "03-3333":
                    sample_3_found = True
                    sequence_files = test_api.get_sequence_files(project_id, sample.sample_name)
                    self.assertEqual(len(sequence_files), 2)
                    res_sequence_file_names = [
                        sequence_files[0]['fileName'],
                        sequence_files[1]['fileName']
                    ]
                    expected_sequence_file_names = [
                        '03-3333_S1_L001_R1_001.fastq.gz',
                        '03-3333_S1_L001_R2_001.fastq.gz'
                    ]
                    self.assertEqual(res_sequence_file_names.sort(), expected_sequence_file_names.sort())

        self.assertEqual(sample_1_found, True)
        self.assertEqual(sample_2_found, True)
        self.assertEqual(sample_3_found, True)

    def test_valid_miseq_with_status_file_already_uploaded(self):
        """
        Test a valid miseq directory for upload from end to end
        We create a status file that indicates the files have already been uploaded,
        Then make sure it does not upload
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=False
        )

        # Write a status file to the upload directory
        directory_status = model.DirectoryStatus(directory=path.join(path_to_module, "fake_ngs_data"))
        directory_status.status = model.DirectoryStatus.COMPLETE
        progress.write_directory_status(directory_status)

        # Do the upload, without force option
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data"), False)

        # Make sure the upload was a failure
        self.assertEqual(upload_result.exit_code, 1)

    def test_invalid_miseq_no_completed_file(self):
        """
        Test a valid miseq directory for upload from end to end
        We create a status file that indicates the files have already been uploaded,
        Then make sure it does not upload
        :return:
        """
        # Set our sample config file to use miseq parser and the correct irida credentials
        self.write_to_config_file(
            client_id=tests_integration.client_id,
            client_secret=tests_integration.client_secret,
            username=tests_integration.username,
            password=tests_integration.password,
            base_url=tests_integration.base_url,
            parser="miseq",
            readonly=False
        )

        # Do the upload, without force option
        upload_result = upload_run_single_entry(path.join(path_to_module, "fake_ngs_data_no_completed_file"), False)

        # Make sure the upload was a failure
        self.assertEqual(upload_result.exit_code, 1)



"""
This file is responsible for Managing an IRIDA instance, including all the parameters associated with testing and
database management.
"""
import requests
from requests import ConnectionError
from urllib.error import URLError
from os import path
from time import time, sleep
import sys
import subprocess
from http import HTTPStatus
from pathlib import Path


class SetupIridaData:

    def __init__(self, base_url, user, password, branch, db_host, db_port):
        """
        :param base_url: url of the IRIDA instance's API
        :param user: default admin username
        :param password: default admin password
        :param branch: the github branch to run the integration tests on (e.g. 'main' or 'development')
        """

        # Login info
        self.base_url = base_url
        self.user = user
        self.password = password
        self.branch = branch

        # Database info
        self.DB_HOST = db_host
        self.DB_PORT = db_port
        self.DB_NAME = "irida_uploader_test"
        self.DB_USERNAME = "test"
        self.DB_PASSWORD = "test"
        self.DB_JDBC_URL = "jdbc:mysql://" + self.DB_HOST + ":" + self.DB_PORT + "/" + self.DB_NAME

        self.TIMEOUT = 600  # seconds

        self.IRIDA_DB_RESET = 'echo '\
            '"drop database if exists ' + self.DB_NAME + ';'\
            'create database ' + self.DB_NAME + ';'\
            '"| mysql -h ' + self.DB_HOST + ' -P ' + self.DB_PORT + ' -u test -ptest'

        root_dir = Path(__file__).parent.absolute()
        sql_file = Path(root_dir, "create_sequencer_client.sql")
        output_files = Path(root_dir, "tmp", "output-files")
        reference_file = Path(root_dir, "tmp", "reference-files")
        sequence_files = Path(root_dir, "tmp", "sequence-files")
        assembly_files = Path(root_dir, "tmp", "assembly-files")

        self.IRIDA_DB_UPDATE = 'mysql -h ' + self.DB_HOST + ' -P ' + self.DB_PORT + ' -u test -ptest ' + \
                               self.DB_NAME + ' < {}'.format(sql_file)

        self.IRIDA_CMD = \
            './gradlew -q clean bootRun -x check ' +\
            '--args=\"' +\
            '--spring.datasource.url={} '.format(self.DB_JDBC_URL) +\
            '--spring.datasource.username={} '.format(self.DB_USERNAME) +\
            '--spring.datasource.password={} '.format(self.DB_PASSWORD) +\
            '--liquibase.update.database.schema=true ' +\
            '--spring.jpa.hibernate.ddl-auto= ' +\
            '--spring.jpa.properties.hibernate.hbm2ddl.import_files= ' +\
            '--sequence.file.base.directory={} '.format(sequence_files) +\
            '--reference.file.base.directory={} '.format(reference_file) +\
            '--output.file.base.directory={} '.format(output_files) +\
            '--assembly.file.base.directory={} '.format(assembly_files) +\
            '--logging.pattern.console=' +\
            '\"'

        self.PATH_TO_MODULE = path.dirname(__file__)
        if len(self.PATH_TO_MODULE) == 0:
            self.PATH_TO_MODULE = "."

        self.SCRIPT_FOLDER = path.join(self.PATH_TO_MODULE, "bash_scripts")
        self.INSTALL_IRIDA_EXEC = path.join(
            self.SCRIPT_FOLDER, "install_irida.sh")

        self.REPO_PATH = path.join(self.PATH_TO_MODULE, "repos")
        self.IRIDA_PATH = path.join(self.REPO_PATH, "irida")

        # Handles the irida subprocess
        self.irida_subprocess = None

    def install_irida(self):
        """
        Installs IRIDA
        :return:
        """
        install_proc = subprocess.Popen(
            [self.INSTALL_IRIDA_EXEC, self.branch], cwd=self.PATH_TO_MODULE)
        proc_res = install_proc.wait()
        if proc_res == 1:  # failed to execute
            sys.exit(1)

    def reset_irida_db(self):
        """
        Deletes the database and creates a new empty one
        :return:
        """
        db_reset_proc = subprocess.Popen(self.IRIDA_DB_RESET, shell=True)
        proc_res = db_reset_proc.wait()

        if proc_res == 1:  # failed to execute
            print("Unable to execute:\n {cmd}".format(cmd=self.IRIDA_DB_RESET))
            sys.exit(1)

    def update_irida_db(self):
        """
        Adds a user and a client for api operations to IRIDA database
        :return:
        """
        # Allow the mysql db time to populate after first start
        # This only is an issue if the machine executing the test is very fast.
        sleep(30)

        db_update_proc = subprocess.Popen(self.IRIDA_DB_UPDATE, shell=True)
        proc_res = db_update_proc.wait()

        if proc_res == 1:  # failed to execute
            print("Unable to execute:\n {cmd}".format(cmd=self.IRIDA_DB_UPDATE))
            sys.exit(1)

    def run_irida(self):
        """
        Runs IRIDA with first initialization
        Waits until IRIDA is up to return
        :return:
        """
        self.irida_subprocess = subprocess.Popen(
            "exec " + self.IRIDA_CMD,  # assign new process image
            stdout=subprocess.PIPE,  # makes subprocess.Popen return pid of command, and not shell (default behaviour)
            cwd=self.IRIDA_PATH,
            shell=True)
        self._wait_until_up()

    def _wait_until_up(self):
        """
        Queries IRIDA until it responds with HTTP 200
        :return:
        """
        start_time = time()
        elapsed = 0
        status_code = -1
        print("Waiting for " + self.base_url)

        while status_code != HTTPStatus.OK and elapsed < self.TIMEOUT:
            try:
                print("Trying to connect...")
                status_code = requests.api.get(self.base_url).status_code
                elapsed = time() - start_time

            except (URLError, ConnectionError):
                sleep(10)
        print("Connected.")

    def stop_irida(self):
        """
        Kill the IRIDA gradlew process
        This will sometimes dump errors into the log, but it is harmless
        :return:
        """
        self.irida_subprocess.kill()
        self.irida_subprocess = None



class IridaResourceError(Exception):
    """
    This error is thrown when the api tries to get a resource route that doesn't exist
    or send data to a resource route that does not exist

    All calls to the api should expect this error
    """

    def __init__(self, message, resource_id=None):
        """Initialize a IridaResourceError.

        Args:
            message: the summary message that's causing the error.
            errors: a more detailed list of errors.
        """
        self._message = message
        self._resource_id = resource_id

    @property
    def message(self):
        return self._message

    @property
    def resource_id(self):
        return self._resource_id

    def __str__(self):
        return self.message



from cerberus import Validator, TypeDefinition

from iridauploader.model.sample import Sample


class Project:

    # Define Sample as a type for validation
    _sample_type = TypeDefinition('sample', (Sample,), ())
    Validator.types_mapping['sample'] = _sample_type

    # schema for sequence file uploading
    uploadable_schema = {
        '_sample_list': {
            'type': 'list',
            'empty': False,  # must have at least 1 sample
            'schema': {'type': 'sample'}},
        '_name': {
            'type': 'string',
            'nullable': True,
            'required': False
        },
        '_description': {
            'type': 'string',
            'nullable': True,
            'required': False
        },
        '_id': {
            'type': 'string',
            'nullable': False,
            'required': True
        }
    }

    # schema for project uploading
    send_project_schema = {
        '_name': {
            'type': 'string',
            'nullable': False,
            'required': True,
            'minlength': 5  # when uploading a new project
        },
        '_description': {
            'type': 'string',
            'nullable': True,
            'required': False
        }
    }

    # project_id is optional because it's not necessary when creating a Project object to send.
    # project_id is the identifier key when getting projects from the API.
    def __init__(self, name=None, sample_list=None, description=None, id=None):
        self._name = name
        if sample_list is None:
            sample_list = []
        self._sample_list = sample_list
        self._description = str(description)
        self._id = str(id)

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def sample_list(self):
        return self._sample_list

    def add_sample(self, sample):
        self._sample_list.append(sample)

    @property
    def description(self):
        return self._description

    def get_uploadable_dict(self):  # formatting for sending to irida when creating a project
        return {"name": self._name,
                "projectDescription": self._description}

    def __str__(self):
        return "ID:" + self._id + " Name: " + self._name + " Description: " + self._description

    def get_dict(self):
        return self.__dict__



"""
This file is responsible for managing the execution of an IRIDA instance and Test Suites
"""
import unittest

from iridauploader.tests_integration.integration_data_setup import SetupIridaData
from iridauploader.tests_integration.test_api_integration import TestApiIntegration
from iridauploader.tests_integration.test_end_to_end import TestEndToEnd

# Modules level variables that can/will be changed when the setup starts
base_url = "http://localhost:8080/api"
username = "jeff"
password = "password1"
client_id = "sequencerClient"
client_secret = "sequencerClientSecret"


def init_irida_setup(branch, db_host, db_port):
    """
    Initializes the Irida setup object
    :param branch: what branch from github to check out
    :param db_host: database host override
    :param db_port: database port override
    :return: SetupIridaData object
    """
    global base_url
    global username
    global password
    global client_id
    global client_secret

    return SetupIridaData(base_url[:base_url.index("/api")], username, password, branch, db_host, db_port)


def create_test_suite():
    """
    Finds all the integration tests and adds them to a test suite
    :return: unittest.TestSuite
    """
    suite = unittest.TestSuite()

    # Include all the classes that contain tests
    test_class_list = [
        TestApiIntegration,
        TestEndToEnd
    ]

    # for each class, find the tests and add them to the test suite
    for test_class in test_class_list:
        for class_method in [*test_class.__dict__]:
            if class_method.startswith("test_"):
                suite.addTest(test_class(class_method))

    return suite


def start(branch="main", db_host="localhost", db_port="3306"):
    """
    Start running the integration tests

    This is the entry point for the integration tests
    :param branch: what branch from github to check out
    :param db_host: database host override
    :param db_port: database port override
    :return:
    """
    exit_code = 0

    # create a handler to manage a headless IRIDA instance
    irida_handler = init_irida_setup(branch, db_host, db_port)
    # Install IRIDA packages
    irida_handler.install_irida()
    # Delete and recreate the database
    irida_handler.reset_irida_db()
    # Launch IRIDA
    # Note: This initializes the database tables
    # Note: This call waits until IRIDA is running
    irida_handler.run_irida()
    # Add data to database that the tests depend on
    irida_handler.update_irida_db()

    # Run tests
    try:
        full_suite = create_test_suite()

        runner = unittest.TextTestRunner()
        res = runner.run(full_suite)
        if not res.wasSuccessful():
            exit_code = 1
    except Exception as e:
        raise e
    finally:
        # Make sure IRIDA is stopped even when an exception is raised
        irida_handler.stop_irida()

    return exit_code



import os
import logging

from configparser import RawConfigParser, NoOptionError
from appdirs import user_config_dir
from collections import namedtuple

_conf_parser = None
_user_config_file = None

_default_user_config_file_path = os.path.join(user_config_dir("irida-uploader"), "config.conf")

# set an override config file in the source code directory, if it exists it will be used instead of getting it from user
_override_config_file_path = os.path.realpath(os.path.join(os.path.dirname(
    os.path.abspath(__file__)), os.path.pardir, "config.conf"))


def set_config_file(config_file):
    """
    Public function to allow other modules to change the config file
    This does not load the settings into the config parser
    :param config_file:
    :return:
    """
    global _user_config_file
    _user_config_file = config_file


def _init_config_parser():
    """
    Creates the RawConfigParser object, adds the fields and fills with default/empty values
    :return:
    """
    global _conf_parser
    # Create a new config parser
    _conf_parser = RawConfigParser()
    # Put all options under a Settings Header
    _conf_parser.add_section("Settings")

    # Create defaults tuple
    SettingsDefault = namedtuple('SettingsDefault', ['setting', 'default_value'])
    default_settings = [SettingsDefault._make(["client_id", ""]),
                        SettingsDefault._make(["client_secret", ""]),
                        SettingsDefault._make(["username", ""]),
                        SettingsDefault._make(["password", ""]),
                        SettingsDefault._make(["base_url", ""]),
                        SettingsDefault._make(["parser", "directory"]),
                        SettingsDefault._make(["readonly", False]),
                        SettingsDefault._make(["delay", 0]),
                        SettingsDefault._make(["timeout", 10]),  # default timeout scale is 10 seconds per mb
                        SettingsDefault._make(["minimum_file_size", 0]),  # default minimum file size in kb
                        SettingsDefault._make(["http_max_retries", 5]),
                        SettingsDefault._make(["http_backoff_factor", 0]),
                        SettingsDefault._make(["log_directory", ""]),
                        ]
    # add defaults to config parser
    for config in default_settings:
        _conf_parser.set("Settings", config.setting, config.default_value)


def _create_new_config_file():
    """
    Tries to create a new config file
    Exit's if the file already exists, or a user config file is already set
    This only gets used to create a fresh default configuration file
    Fills the new file with defaults
    :return:
    """
    global _user_config_file
    if _user_config_file is not None:
        logging.error("Trying to create new config file when config file as already been set. Exiting")
        return

    _user_config_file = _default_user_config_file_path

    # Create path to file if it doesn't exit yet
    if not os.path.exists(os.path.dirname(_user_config_file)):
        os.makedirs(os.path.dirname(_user_config_file))
    # return if the file already exists
    if os.path.exists(_user_config_file):
        logging.error("File already exits, cannot create a new file")
        return

    # init the config parser with default values
    global _conf_parser
    with open(_user_config_file, 'w') as config_file:
        _conf_parser.write(config_file)

    logging.info("Config File Created: " + str(_user_config_file))
    logging.info("Please edit your config file to connect to IRIDA")


def _load_config_from_file():
    """
    Verifies file exists, and loads data from file into the config parser
    :return:
    """
    global _user_config_file

    if not os.path.exists(_user_config_file):
        logging.error("Config File {} does not exist, exiting".format(_user_config_file))
        exit(1)

    try:
        read_config_file_into_parser(_user_config_file)
    except Exception:
        logging.error("Config file {} is not valid.".format(_user_config_file))
        exit(1)

    logging.info("Config loaded from file {}".format(_user_config_file))


def read_config_file_into_parser(config_file_path):
    """
    Read the config file into memory using the global _config_parser instance
    :param config_file_path:
    :return:
    """
    try:
        global _conf_parser
        _conf_parser.read(config_file_path)
    except Exception as e:
        logging.warning("Error occurred when trying to load config file")
        raise e


def set_config_options(client_id=None,
                       client_secret=None,
                       username=None,
                       password=None,
                       base_url=None,
                       parser=None,
                       readonly=None,
                       delay=None,
                       timeout=None,
                       minimum_file_size=None,
                       http_max_retries=None,
                       http_backoff_factor=None,
                       log_directory=None):
    """
    Updates the config options for all not None parameters
    :param client_id:
    :param client_secret:
    :param username:
    :param password:
    :param base_url:
    :param parser:
    :param readonly:
    :param delay:
    :param timeout:
    :param minimum_file_size:
    :param http_max_retries:
    :param http_backoff_factor:
    :param log_directory:
    :return:
    """
    global _conf_parser
    if _conf_parser is None:
        logging.error("Config Parser has not been init, exiting")
        exit(1)

    if client_id:
        logging.debug("Setting 'client_id' config to {}".format(client_id))
        _update_config_option("client_id", client_id)
    if client_secret:
        logging.debug("Setting 'client_secret' config to {}".format(client_secret))
        _update_config_option("client_secret", client_secret)
    if username:
        logging.debug("Setting 'username' config to {}".format(username))
        _update_config_option("username", username)
    if password:
        logging.debug("Setting 'password' config")
        _update_config_option("password", password)
    if base_url:
        logging.debug("Setting 'base_url' config to {}".format(base_url))
        _update_config_option("base_url", base_url)
    if parser:
        logging.debug("Setting 'parser' config to {}".format(parser))
        _update_config_option("parser", parser)
    # since readonly is a bool and not a string, we need to check that is is not None, not just not True
    if readonly is not None:
        logging.debug("Setting 'readonly' config to {}".format(readonly))
        _update_config_option("readonly", readonly)
    if delay is not None:
        # delay is always an int
        logging.debug("Setting 'delay' config to {}".format(delay))
        _update_config_option('delay', delay)
    if timeout is not None:
        # timeout is always an int
        logging.debug("Setting 'timeout' config to {}".format(timeout))
        _update_config_option('timeout', timeout)
    if minimum_file_size is not None:
        # minimum_file_size is always an int
        logging.debug("Setting 'minimum_file_size' config to {}".format(minimum_file_size))
        _update_config_option('minimum_file_size', minimum_file_size)
    if http_max_retries is not None:
        # http_max_retries is always an int
        logging.debug("Setting 'http_max_retries' config to {}".format(http_max_retries))
        _update_config_option('http_max_retries', http_max_retries)
    if http_backoff_factor is not None:
        # http_backoff_factor is always a float
        logging.debug("Setting 'http_backoff_factor' config to {}".format(http_backoff_factor))
        _update_config_option('http_backoff_factor', http_backoff_factor)
    if log_directory:
        # log_directory is always a str
        logging.debug("Setting 'log_directory' config to {}".format(log_directory))
        _update_config_option('log_directory', log_directory)


def setup():
    """
    Initialize and setup components for the configuration file
    Creates a new conf parser, and loads from user_config_file or default config file
    If the default file is used but doesn't exist, it will be created and the program will exit
    :return:
    """
    global _conf_parser
    global _user_config_file
    global _default_user_config_file_path
    global _override_config_file_path

    _init_config_parser()
    # If a config file is set, inform user
    if _user_config_file:
        logging.info("Config file set to {}".format(_user_config_file))
    # if the override config file exists, use it over the default
    elif os.path.exists(_override_config_file_path):
        logging.info("Override config file found. Using file: {}".format(_override_config_file_path))
        _user_config_file = _override_config_file_path
    # If a config file was not set, create a new file
    elif not os.path.exists(_default_user_config_file_path):
        logging.info("No config file found, creating a new file {}".format(_default_user_config_file_path))
        _create_new_config_file()
    # Use the default file as the user config file
    else:
        logging.info("Using default config file {}".format(_default_user_config_file_path))
        _user_config_file = _default_user_config_file_path

    # Load from file
    _load_config_from_file()


def read_config_option(key, expected_type=None, default_value=None):
    """Read the specified value from the configuration file.

    Args:
        key: the name of the key to read from the config file.
        expected_type: read the config option as the specified type (if specified)
        default_value: if the key doesn't exist, just return the default value.
            If the default value is not specified, the function will throw whatever
            error was raised by the configuration parser
    """
    logging.debug("Reading config option {} with expected type {}".format(key, expected_type))

    global _conf_parser
    try:
        if not expected_type:
            value = _conf_parser.get("Settings", key)
            logging.debug("Got configuration for key {}: {}".format(key, value))
            return _conf_parser.get("Settings", key)
        elif expected_type is int:
            res = _conf_parser.get("Settings", key)
            logging.debug("Got configuration for key {}: {}".format(key, res))
            # Return int, or string evaluated to int, or NameError exception otherwise
            if type(res) is int:
                return res
            elif type(res) is str:
                try:
                    return int(res)
                except Exception:
                    error_msg = f"Config file field '{key}' expected int but instead got '{res}'"
                    logging.error(error_msg)
                    raise NameError(error_msg)
        elif expected_type is float:
            res = _conf_parser.get("Settings", key)
            logging.debug("Got configuration for key {}: {}".format(key, res))
            # Return int, or string evaluated to int, or NameError exception otherwise
            if type(res) is float:
                return res
            elif type(res) is str:
                try:
                    return float(res)
                except Exception:
                    error_msg = f"Config file field '{key}' expected float but instead got '{res}'"
                    logging.error(error_msg)
                    raise NameError(error_msg)
        elif expected_type is bool:
            res = _conf_parser.get("Settings", key)
            logging.debug("Got configuration for key {}: {}".format(key, res))
            # Return bool, or string evaluated to bool, or NameError exception otherwise
            if type(res) is bool:
                return res
            elif type(res) is str:
                return _eval_boolean(res, key)
            else:
                raise NameError
    except (ValueError, NameError, NoOptionError):
        if default_value is not None:
            return default_value
        else:
            raise


def _eval_boolean(string, field):
    """
    Converts string to bool value, accepts any case permutation.
    If non true/false is given, displays error to user and raises to exit
    :param string: string to evaluate as a boolean
    :param field: config field the key originated from for logging and error message
    :return: Boolean
    """
    caps = string.upper()
    if caps == "TRUE":
        logging.debug("Evaluated as True")
        return True
    elif caps == "FALSE":
        logging.debug("Evaluated as False")
        return False
    else:
        error_msg = "Config file field '{}' expected 'True' or 'False' but instead got '{}'".format(field, string)
        logging.error(error_msg)
        raise NameError(error_msg)


def _update_config_option(field_name, field_value):
    """
    Sets an option in the config parser (Does not write to file)
    :param field_name: Field to change the value of
    :param field_value: Value to change to
    :return:
    """
    global _conf_parser
    logging.debug("Setting [" + field_name + "] to [" + str(field_value) + "]")
    _conf_parser.set("Settings", field_name, field_value)


def write_config_options_to_file():
    """
    Writes all options in the config parser to config file

    :return: None
    """
    global _conf_parser
    global _user_config_file
    with open(_user_config_file, 'w') as c_file:
        _conf_parser.write(c_file)



import os.path

import iridauploader.model as model
import iridauploader.config as config
from iridauploader.parsers.exceptions.file_size_error import FileSizeError


def validate_file_size_minimum(sequencing_run):
    """
    Validate the files in a SequencingRun object have the minimum file size requirement from the config

    :param sequencing_run: SequencingRun object to validate
    :return: ValidationResult object with list of errors if any
    """

    minimum_file_size = config.read_config_option("minimum_file_size", int, 0)

    validation_result = model.ValidationResult()

    for p in sequencing_run.project_list:
        for s in p.sample_list:
            # do validation of file size
            if not _file_size_is_valid(s.sequence_file, minimum_file_size):
                error_msg = "File size for sample `{}`is smaller than configured minimum of `{} KB`. " \
                            "Please verify your data.".format(s.sample_name, minimum_file_size)
                validation_result.add_error(FileSizeError(error_msg, s.sequence_file))

    return validation_result


def _file_size_is_valid(sequence_file, minimum_file_size):
    """
    Performs a os.path.getsize operation on the sequence file(s) on the sequence_file object
    if any files are less than the file minimum, return false, else return true

    :param sequence_file: SequenceFile object
    :param minimum_file_size: integer representing filesize in KB
    :return: boolean
    """
    for file_name in sequence_file.file_list:
        file_size_kb = os.path.getsize(file_name) / 1024
        if file_size_kb <= minimum_file_size:
            return False

    return True



import re
from os import path, walk
from collections import OrderedDict
from copy import deepcopy
import logging

import iridauploader.model as model
from iridauploader.parsers import exceptions
from iridauploader.parsers import common


def parse_metadata(sample_sheet_file):

    """
    Parse all lines under [Header], [Reads] and [Settings] in .csv file
    Lines under [Reads] are stored in a list with key name "readLengths"
    All other key names are translated according to the
        metadata_key_translation_dict

    arguments:
            sample_sheet_file -- path to SampleSheet.csv

    returns a dictionary containing the parsed key:pair values from .csv file
    """

    metadata_dict = {"readLengths": []}

    csv_reader = common.get_csv_reader(sample_sheet_file)

    metadata_key_translation_dict = {
        'Assay': 'assay',
        'Description': 'description',
        'Application': 'application',
        'Investigator Name': 'investigatorName',
        'Adapter': 'adapter',
        'AdapterRead2': 'adapterread2',
        'Workflow': 'workflow',
        'ReverseComplement': 'reversecomplement',
        'IEMFileVersion': 'iemfileversion',
        'Date': 'date',
        'Experiment Name': 'experimentName',
        'Chemistry': 'chemistry',
        'Project Name': 'projectName'
    }

    section = None

    for line in csv_reader:
        if "[Header]" in line or "[Settings]" in line:
            section = "header"
            continue
        elif "[Reads]" in line:
            section = "reads"
            continue
        elif "[Data]" in line:
            break
        elif line and line[0].startswith("["):
            section = "unknown"
            continue

        if not line or not line[0]:
            continue

        if not section:
            logging.debug("Sample sheet is missing important sections: no sections were found")
            raise exceptions.SampleSheetError("Sample sheet is missing important sections: no sections were found.",
                                              sample_sheet_file)
        elif section == "header":
            try:
                key_name = metadata_key_translation_dict[line[0]]
                metadata_dict[key_name] = line[1]
            except KeyError:
                logging.debug("Unexpected key in header: [{}]".format(line[0]))
        elif section == "reads":
            metadata_dict["readLengths"].append(line[0])

    # currently sends just the larger readLengths
    if len(metadata_dict["readLengths"]) > 0:
        if len(metadata_dict["readLengths"]) == 2:
            metadata_dict["layoutType"] = "PAIRED_END"
        else:
            metadata_dict["layoutType"] = "SINGLE_END"
        metadata_dict["readLengths"] = max(metadata_dict["readLengths"])
    else:
        # this is an exceptional case, you can't have no read lengths!
        logging.debug("The sample sheet is missing important sections: no [Reads] section found.")
        raise exceptions.SampleSheetError("The sample sheet is missing important sections: no [Reads] section found.",
                                          sample_sheet_file)

    return metadata_dict


def build_sequencing_run_from_samples(sample_sheet_file, metadata, sequencing_run_type, strict_matching=False):
    """
    Create a SequencingRun object with full project/sample/sequence_file structure

    :param sample_sheet_file:
    :param metadata:
    :param sequencing_run_type:
    :return: SequencingRun
    """
    sample_list = _parse_sample_list(sample_sheet_file, strict_matching)

    logging.debug("Building SequencingRun from parsed data")

    # create list of projects and add samples to appropriate project
    project_list = []
    for sample in sample_list:
        project = None
        for p in project_list:
            if sample.get('sample_project') == p.id:
                project = p
        if project is None:
            project = model.Project(id=sample.get('sample_project'))
            project_list.append(project)

        project.add_sample(sample)

    sequence_run = model.SequencingRun(metadata=metadata,
                                       project_list=project_list,
                                       sequencing_run_type=sequencing_run_type)
    logging.debug("SequencingRun built")
    return sequence_run


def _parse_sample_list(sample_sheet_file, strict_matching=False):
    """
    Creates a list of all samples in the sample_sheet_file, with accompanying data/metadata

    :param sample_sheet_file:
    :return: list of samples
    """
    sample_list = _parse_samples(sample_sheet_file)
    sample_sheet_dir = path.dirname(sample_sheet_file)
    base_data_dir = path.join(sample_sheet_dir, "Data", "Intensities", "BaseCalls")
    project_dir_list = next(walk(base_data_dir))[1]  # Get the list of project directories that contain sample files

    for sample in sample_list:

        project_directory = sample.get('sample_project')
        if project_directory not in project_dir_list:
            # The project number in the sample sheet does not match a project folder in the run directory
            raise exceptions.SequenceFileError(
                "The uploader was unable to find the directory '{}' in '{}'. "
                "Please verify your SampleSheet Sample_Project matches the directory structure".format(
                    project_directory, base_data_dir
                )
            )
        project_data_dir = path.join(base_data_dir, project_directory)
        # Create a file list of the data directory, only hit the os once
        data_dir_file_list = next(walk(project_data_dir))[2]

        properties_dict = _parse_out_sequence_file(sample)
        file_pattern = "{sample_name}_S(\\S+)_R(\\d+)_(\\S*)\\.fastq.*$".format(
            sample_name=re.escape(sample.sample_name))
        if strict_matching:
            file_pattern = "^" + file_pattern
        logging.info("Looking for files with pattern {}".format(file_pattern))
        regex = re.compile(file_pattern)
        pf_list = list(filter(regex.search, data_dir_file_list))

        if not pf_list:
            # we didn't find anything
            raise exceptions.SequenceFileError(
                ("The uploader was unable to find an files with a file name that ends with "
                 ".fastq.gz for the sample in your sample sheet with name {} in the directory {}. ").format(
                    sample.sample_name, project_data_dir))

        # List of files may be invalid if directory searching in has been modified by user
        if not _validate_pf_list(pf_list):
            raise exceptions.SequenceFileError(
                ("The following file list {} found in the directory {} is invalid. "
                 "Please verify the folder containing the sequence files matches the SampleSheet file").format(
                    pf_list, project_data_dir))

        # Add the dir to each file to create the full path
        for i in range(len(pf_list)):
            pf_list[i] = path.join(project_data_dir, pf_list[i])

        sq = model.SequenceFile(file_list=pf_list, properties_dict=properties_dict)
        sample.sequence_file = deepcopy(sq)

    return sample_list


def _validate_pf_list(file_list):
    """
    Checks if list of files is valid:

    arguments:
            file_list -- list of file names that are grouped together

    returns:
            True: 1 file in list
            True: 2 files in list, where one contains `R1` in the correct position and the other contains `R2`
            False: Number of files <1 or >2, or 2 files do not contain `R1`/`R2` correctly
    """
    if len(file_list) < 1:  # Invalid
        return False
    if len(file_list) > 2:  # We should never expect more than 2 files, a forward & backwards read
        return False
    elif len(file_list) == 1:  # single read, valid
        return True
    else:
        try:
            # check if one file is R1 and other is R2
            regex_filter = ".*_S\\d+_R(\\d+)_\\S+\\.fastq.*$"
            n1 = int(re.search(regex_filter, file_list[0]).group(1))
            n2 = int(re.search(regex_filter, file_list[1]).group(1))
            return (n1 != n2) and (n1 == 1 or n1 == 2) and (n2 == 1 or n2 == 2)
        except AttributeError:  # Attribute error means the file had invalid text in it
            return False


def _parse_samples(sample_sheet_file):

    """
    Parse all the lines under "[Data]" in .csv file
    Keys in sample_key_translation_dict have their values changed for
        uploading to REST API
    All other keys keep the same name that they have in .csv file

    arguments:
            sample_sheet_file -- path to SampleSheet.csv

    returns	a list containing Sample objects that have been created by a
        dictionary from the parsed out key:pair values from .csv file
    """

    logging.info("Reading data from sample sheet {}".format(sample_sheet_file))

    csv_reader = common.get_csv_reader(sample_sheet_file)
    # start with an ordered dictionary so that keys are ordered in the same
    # way that they are inserted.
    sample_dict = OrderedDict()
    sample_list = []

    sample_key_translation_dict = {
        'Sample_Name': 'sequencer_sample_name',
        'Description': 'description',
        'Sample_ID': 'sampleName',
        'Sample_Project': 'sample_project'
    }

    _parse_samples.sample_key_translation_dict = sample_key_translation_dict

    # initilize dictionary keys from first line (data headers/attributes)
    set_attributes = False
    for line in csv_reader:

        if set_attributes:
            for item in line:

                if item in sample_key_translation_dict:
                    key_name = sample_key_translation_dict[item]
                else:
                    key_name = item

                sample_dict[key_name] = ""

            break

        if "[Data]" in line:
            set_attributes = True

    # fill in values for keys. line is currently below the [Data] headers
    for sample_number, line in enumerate(csv_reader):

        if len(sample_dict.keys()) != len(line):
            """
            if there is one more Data header compared to the length of
            data values then add an empty string to the end of data values
            i.e the Description will be empty string
            assumes the last Data header is going to be the Description
            this handles the case where the last trailing comma is trimmed

            Shaun said this issue may come up when a user edits the
            SampleSheet from within the NextSeq software
            """
            if len(sample_dict.keys()) - len(line) == 1:
                line.append("")
            else:
                raise exceptions.SampleSheetError(
                    ("Your sample sheet is malformed. Expected to find {} "
                     "columns in the [Data] section, but only found {} columns "
                     "for line {}.".format(len(sample_dict.keys()), len(line), line)),
                    sample_sheet_file
                )

        for index, key in enumerate(sample_dict.keys()):
            sample_dict[key] = line[index].strip()  # assumes values are never empty

        new_sample_dict = deepcopy(sample_dict)
        new_sample_name = new_sample_dict['sampleName']
        new_sample_desc = new_sample_dict['description']
        del new_sample_dict['sampleName']
        del new_sample_dict['description']

        sample = model.Sample(
            sample_name=new_sample_name,
            description=new_sample_desc,
            sample_number=sample_number + 1,
            samp_dict=new_sample_dict)
        sample_list.append(sample)

    return sample_list


def _parse_out_sequence_file(sample):

    """
    Removes keys in argument sample that are not in sample_keys and
        stores them in sequence_file_dict

    arguments:
            sample -- Sample object
            the dictionary inside the Sample object is changed

    returns a dictionary containing keys not in sample_keys to be used to
        create a SequenceFile object
    """

    sample_keys = ["sequencer_sample_name", "description",
                   "sampleName", "sample_project"]
    sequence_file_dict = {}
    sample_dict = sample.get_uploadable_dict()
    for key in list(sample_dict.keys()):
        if key not in sample_keys:
            sequence_file_dict[key] = sample_dict[key]

    return sequence_file_dict



"""
A SequencingRun contains all required information from a run

project_list: list of Project objects
metadata: dict of key value pairs for additional data
    the metadata dict requires the key `layoutType` with one of ['PAIRED_END', 'SINGLE_END']
assemblies: boolean, True if uploading assemblies
sequencing_run_type: string for path uploading to.
"""
from cerberus import Validator, TypeDefinition

from iridauploader.model.project import Project


class SequencingRun:

    # Define Project as a type for validation
    _project_type = TypeDefinition('project', (Project,), ())
    Validator.types_mapping['project'] = _project_type

    uploadable_schema = {
        '_project_list': {
            'type': 'list',
            'empty': False,  # must have at least 1 project
            'schema': {'type': 'project'}
        },
        '_metadata': {
            'type': 'dict',
            'schema': {
                'layoutType': {
                    'type': 'string',
                    'required': True,
                    'allowed': ['PAIRED_END', 'SINGLE_END']
                }
            }
        },
        '_sequencing_run_type': {
            'type': 'string',
            'required': True
        }
    }

    def __init__(self, metadata, project_list, sequencing_run_type):
        self._project_list = project_list
        self._metadata = metadata
        self._sequencing_run_type = sequencing_run_type

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, metadata_dict):
        self._metadata = metadata_dict

    @property
    def project_list(self):
        return self._project_list

    @project_list.setter
    def project_list(self, p_list):
        self._project_list = p_list

    @property
    def sequencing_run_type(self):
        return self._sequencing_run_type

    @sequencing_run_type.setter
    def sequencing_run_type(self, sequencing_run_type):
        self._sequencing_run_type = sequencing_run_type

    def get_dict(self):
        return self.__dict__

    def is_paired_end(self):
        """
        Checks the metadata field to see if run is paired end or single end
        :return: boolean
        """
        layout_type = self.metadata['layoutType']
        return True if layout_type == "PAIRED_END" else False



import logging
import os

import iridauploader.model as model
import iridauploader.progress as progress

from iridauploader.parsers import BaseParser
from iridauploader.parsers import exceptions
from iridauploader.parsers import common
from iridauploader.parsers.nextseq import sample_parser, validation


class Parser(BaseParser):

    UPLOAD_COMPLETE_FILE_NAME = 'RTAComplete.txt'

    def __init__(self, parser_type_name='miseq', sample_sheet_override=None, additional_required_files=None, strict_sample_name_matching=False):
        """
        Initialize the Parser
        :param parser_type_name: string to be included in metadata of sequencing run for type identification in IRIDA
        """
        if sample_sheet_override is not None:
            self.SAMPLE_SHEET_FILE_NAME = sample_sheet_override
        else:
            self.SAMPLE_SHEET_FILE_NAME = 'SampleSheet.csv'

        req_files = [self.SAMPLE_SHEET_FILE_NAME, Parser.UPLOAD_COMPLETE_FILE_NAME]
        if additional_required_files is not None:
            for f in additional_required_files:
                req_files.append(f)

        self.strict_sample_name_matching = strict_sample_name_matching

        super().__init__(
            parser_type_name=parser_type_name,
            required_file_list=req_files
        )
        logging.warning("NOTE: If bcl2fastq has not finished, run may return as invalid, "
                        "or incomplete files could be uploaded!")

    def find_runs(self, directory):
        """
        find a list of run directories in the directory given

        :param directory:
        :return: list of DirectoryStatus objects
        """
        logging.info("looking for runs in {}".format(directory))

        runs = []
        directory_list = common.find_directory_list(directory)
        for d in directory_list:
            runs.append(progress.get_directory_status(d, self.get_required_file_list()))

        return runs

    def find_single_run(self, directory):
        """
        Find a run in the base directory given

        :param directory:
        :return: DirectoryStatus object
        """
        logging.info("looking for run in {}".format(directory))

        return progress.get_directory_status(directory, self.get_required_file_list())

    def get_sample_sheet(self, directory):
        """
        gets the sample sheet file path from a given run directory

        :param directory:
        :return:
        """
        logging.info("Looking for sample sheet in {}".format(directory))

        # Checks if we can access to the given directory, return empty and log a warning if we cannot.
        if common.cannot_read_directory(directory):
            logging.error(("The directory is not accessible, can not parse samples from this directory {}"
                           "".format(directory), directory))
            raise exceptions.DirectoryError("The directory is not accessible, "
                                            "can not parse samples from this directory {}".format(directory), directory)

        sample_sheet_file_name = self.SAMPLE_SHEET_FILE_NAME
        file_list = next(os.walk(directory))[2]  # Gets the list of files in the directory
        if sample_sheet_file_name not in file_list:
            logging.error("No sample sheet file in the NextSeq format found")
            raise exceptions.DirectoryError("The directory {} has no sample sheet file in the NextSeq format"
                                            " with the name {}"
                                            .format(directory, sample_sheet_file_name), directory)
        else:
            logging.debug("Sample sheet found")
            return os.path.join(directory, sample_sheet_file_name)

    def get_sequencing_run(self, sample_sheet):
        """
        Does local validation on the integrity of the run directory / sample sheet

        Throws a ValidationError with a validation result attached if it cannot make a sequencing run

        :param sample_sheet:
        :return: SequencingRun
        """

        # Try to get the sample sheet, validate that the sample sheet is valid
        validation_result = validation.validate_sample_sheet(sample_sheet)
        if not validation_result.is_valid():
            logging.error("Errors occurred while getting sample sheet")
            raise exceptions.ValidationError("Errors occurred while getting sample sheet", validation_result)

        # Try to parse the meta data from the sample sheet, throw validation error if errors occur
        validation_result = model.ValidationResult()
        try:
            run_metadata = sample_parser.parse_metadata(sample_sheet)
        except exceptions.SampleSheetError as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while parsing metadata")
            raise exceptions.ValidationError("Errors occurred while parsing metadata", validation_result)

        # Try to build sequencing run from sample sheet & meta data, raise validation error if errors occur
        try:
            sequencing_run = sample_parser.build_sequencing_run_from_samples(sample_sheet,
                                                                             run_metadata,
                                                                             self.get_parser_type_name(), self.strict_sample_name_matching)
        except exceptions.SequenceFileError as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while building sequence run from sample sheet")
            raise exceptions.ValidationError("Errors occurred while building sequence run from sample sheet",
                                             validation_result)

        return sequencing_run



from iridauploader.parsers import exceptions
from iridauploader.parsers import common
import iridauploader.model as model


def validate_sample_sheet(sample_sheet_file):

    """
    Checks if the given sample_sheet_file can be parsed
    Requires [Header] because it contains Workflow
    Requires [Data] for creating Sample objects and requires
        Sample_ID, Sample_Name, Sample_Project and Description table headers

    arguments:
            sample_sheet_file -- path to SampleSheet.csv

    returns ValidationResult object - stores list of string error messages
    """

    csv_reader = common.get_csv_reader(sample_sheet_file)

    v_res = model.ValidationResult()

    all_data_headers_found = False
    data_sect_found = False
    check_data_headers = False

    # status of required data headers
    found_data_headers = {
        "Sample_Name": False,
        "Project_ID": False,
        "File_Forward": False,
        "File_Reverse": False}

    for line in csv_reader:

        if "[Data]" in line:
            data_sect_found = True
            check_data_headers = True  # next line contains data headers

        elif check_data_headers:
            for data_header in found_data_headers.keys():
                if data_header in line:
                    found_data_headers[data_header] = True

            # if all required dataHeaders are found
            if all(found_data_headers.values()):
                all_data_headers_found = True

            check_data_headers = False

    if not all([data_sect_found, all_data_headers_found]):

        if data_sect_found is False:
            v_res.add_error(exceptions.SampleSheetError("[Data] section not found in SampleSheet", sample_sheet_file))

        if all_data_headers_found is False:
            missing_str = ""
            for data_header in found_data_headers:
                if found_data_headers[data_header] is False:
                    missing_str = missing_str + data_header + ", "

            missing_str = missing_str[:-2]  # remove last ", "
            v_res.add_error(exceptions.SampleSheetError("Missing required data header(s): " + missing_str,
                                                        sample_sheet_file))

    return v_res



#!/usr/bin/env python3
"""
This file is the entry point for the command line interface for both bash and windows

It uses argparse, getpass, as well as os to keep the interface operating system agnostic
"""

import argparse
import getpass
import os
import textwrap

from iridauploader import VERSION_NUMBER
import iridauploader.config as config
import iridauploader.core as core
from iridauploader.api import UPLOAD_MODES
from iridauploader.parsers import supported_parsers

DESCRIPTION = textwrap.dedent('''
This program parses sequencing runs and uploads them to IRIDA.

required arguments:
  -d DIRECTORY, --directory DIRECTORY
                        Location of sequencing run to upload.
                        Directory must be writable.
''')


def init_argparser():
    # Set up an argument parser. We are using defaults to stay consistent with other software.
    # description gets added to the usage statements
    argument_parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=DESCRIPTION,
        prog="irida-uploader -d DIRECTORY",
        epilog='-c* options can be used without a parameter to prompt for input.')
    # Our main argument. It is required or else an error will be thrown when the program is run
    # Normally we would use a positional argument, but because of our 1 or None config overrides it makes sense to use
    # a optional argument, with the required set to True. We have to suppress the help on the argument, and add it's
    # help information to the formatted description text of the parser for the argument to be shown as required when
    # --help is used.
    argument_parser.add_argument('-d', '--directory',
                                 action='store',
                                 required=True,
                                 help=argparse.SUPPRESS)
    # help='Location of sequencing run to upload. Directory must be writable.')
    # Add the version argument
    argument_parser.add_argument('--version',
                                 action='version', version='IRIDA Uploader {}'.format(VERSION_NUMBER))
    # Optional argument, for using an alternative config file.
    argument_parser.add_argument('-c', '--config',
                                 action='store',
                                 help='Path to an alternative configuration file. '
                                      'This overrides the default config file in the config directory')
    # Optional argument, Force uploading a run even if a non new status file exists
    argument_parser.add_argument('-f', '--force',
                                 action='store_true',  # This line makes it not parse a variable
                                 help='Uploader will ignore the status file, '
                                      'and try to upload even when a run is in non new status.')
    # Optional argument, Force uploading a run even if a non new status file exists
    argument_parser.add_argument('-p', '--continue_partial',
                                 action='store_true',  # This line makes it not parse a variable
                                 help='Uploader will continue an existing upload run when upload status is partial.')
    # Optional argument, Upload all sequencing runs in a directory of runs
    argument_parser.add_argument('-b', '--batch',
                                 action='store_true',  # This line makes it not parse a variable
                                 help='Uploader will expect a directory containing a sequencing run directories, '
                                      'and upload in batch. '
                                      'The list of runs is generated at start time '
                                      '(Runs added to directory mid upload will not be uploaded).')
    # Optional argument, Upload mode
    argument_parser.add_argument('-u', '--upload_mode',
                                 action='store',
                                 help='Choose which upload mode to use. '
                                      'Supported modes: ' + str(UPLOAD_MODES))

    # Optional arguments for overriding config file settings
    # Explanation:
    #   nargs='?', const=True, default=False,
    #   Allows zero or one parameters
    #   If the argument is not given:                      the value will be False           (indicates load from file)
    #   If the argument is given, and no parameter given:  the value will be True            (prompt user for input)
    #   If the argument is given, and parameter is given:  the value will be the parameter   (used to override config)
    argument_parser.add_argument('-ci', '--config_client_id', action='store', nargs='?', const=True, default=False,
                                 help='Override the "client_id" field in config file. '
                                      'This is for the uploader client created in IRIDA, '
                                      'used to handle the data upload')
    argument_parser.add_argument('-cs', '--config_client_secret', action='store', nargs='?', const=True, default=False,
                                 help='Override "client_secret" in config file. '
                                      'This is for the uploader client created in IRIDA, '
                                      'used to handle the data upload')
    argument_parser.add_argument('-cu', '--config_username', action='store', nargs='?', const=True, default=False,
                                 help='Override "username" in config file. This is your IRIDA account username.')
    argument_parser.add_argument('-cp', '--config_password', action='store', nargs='?', const=True, default=False,
                                 help='Override "password" in config file. This corresponds to your IRIDA account.')
    argument_parser.add_argument('-cb', '--config_base_url', action='store', nargs='?', const=True, default=False,
                                 help='Override "base_url" in config file. The api url for your irida instance '
                                      '(example: https://my.irida.server/api/)')
    argument_parser.add_argument('-cr', '--config_parser', action='store', nargs='?', const=True, default=False,
                                 help='Override "parser" in config file. '
                                      'Choose the type of sequencer data that is being uploaded. '
                                      'Supported parsers: ' + str(supported_parsers))
    argument_parser.add_argument('-r', '--readonly',
                                 action='store_true',  # This line makes it not parse a variable
                                 help='Upload in Read Only mode, does not write status or log file to run directory.')
    argument_parser.add_argument('-cd', '--delay', action='store', nargs='?', const=True, default=False,
                                 help='Accepts an Integer for minutes to delay. When set, new runs will have their  '
                                      'status set to delayed. When uploading a run with the delayed status, the run '
                                      'will only upload if the given amount of minutes has passes since it was set to '
                                      'delayed. Default = 0: When set to 0, runs will not be given delayed status.')
    argument_parser.add_argument('-ct', '--config_timeout', action='store', nargs='?', const=True, default=False,
                                 help='Accepts an Integer for the expected transfer time in seconds per MB. '
                                      'Default is 10 second for every MB of data to transfer (100kb/s). Increasing this'
                                      ' number can reduce timeout errors when your transfer speed is very slow.')
    argument_parser.add_argument('-fs', '--minimum_file_size', action='store', nargs='?', const=True, default=False,
                                 help='Accepts an Integer for the minimum file size in KB. Default is 0 KB. Files that '
                                      'are too small will appear as an error during run validation.')
    return argument_parser


def _set_config_override(args):
    """
    Check the config override arguments and override
    :param args: list of args from parseargs
    :return:
    """
    client_id = None
    client_secret = None
    username = None
    password = None
    base_url = None
    parser = None
    readonly = None
    delay = None
    timeout = None
    minimum_file_size = None

    if args.config_client_id is True:
        print("Enter Client ID:")
        client_id = input()
    elif args.config_client_id is not False:
        client_id = args.config_client_id

    if args.config_client_secret is True:
        print("Enter Client Secret:")
        # getpass blanks out the secret entry
        client_secret = getpass.getpass()
    elif args.config_client_secret is not False:
        client_secret = args.config_client_secret

    if args.config_username is True:
        print("Enter Username:")
        username = input()
    elif args.config_username is not False:
        username = args.config_username

    if args.config_password is True:
        print("Enter Password:")
        # getpass blanks out password entry
        password = getpass.getpass()
    elif args.config_password is not False:
        password = args.config_password

    if args.config_base_url is True:
        print("Enter Base IRIDA URL (format: http://my.irida.server/api/):")
        base_url = input()
    elif args.config_base_url is not False:
        base_url = args.config_base_url

    if args.config_parser is True:
        print("Enter Parser to use:")
        parser = input()
    elif args.config_parser is not False:
        parser = args.config_parser

    if args.readonly is True:
        readonly = args.readonly

    if args.delay is True:
        print("Enter delay in minutes (Integer):")
        delay = int(input())
    elif args.delay is not False:
        delay = int(args.delay)

    if args.config_timeout is True:
        print("Enter timeout per MB in seconds (Integer):")
        timeout = int(input())
    elif args.config_timeout is not False:
        timeout = int(args.config_timeout)

    if args.minimum_file_size is True:
        print("Enter minimum file size in KB (Integer):")
        minimum_file_size = int(input())
    elif args.minimum_file_size is not False:
        minimum_file_size = int(args.minimum_file_size)

    config.set_config_options(client_id=client_id,
                              client_secret=client_secret,
                              username=username,
                              password=password,
                              base_url=base_url,
                              parser=parser,
                              readonly=readonly,
                              delay=delay,
                              timeout=timeout,
                              minimum_file_size=minimum_file_size)


def _config_uploader(args):
    """
    Sets up config settings for command line uploading
    :param args:
    :return:
    """
    # If a config file is passed in, set it before starting upload
    if args.config:
        config.set_config_file(args.config)
        # config.user_config_file = args.config
    # Init config
    config.setup()
    # Override with any passed in options
    _set_config_override(args)


def main():
    # Parse the arguments passed from the command line and start the upload
    argument_parser = init_argparser()
    args = argument_parser.parse_args()
    # Setup config options
    _config_uploader(args)

    # Verify directory is readable before upload
    if not os.access(args.directory, os.R_OK):  # Cannot access upload directory
        print("ERROR! Specified directory is not readable: {}".format(args.directory))
        return 1

    # Verify force and continue are not both active
    if args.force and args.continue_partial:
        print("ERROR! --force and --continue_partial are mutually exclusive. "
              "To continue a partial upload use --continue_partial. "
              "To upload all samples from the beginning use --force")
        return 1

    # Start Upload
    if args.batch:
        return upload_batch(args.directory, args.force, args.upload_mode, args.continue_partial)
    else:
        return upload(args.directory, args.force, args.upload_mode, args.continue_partial)


def upload(run_directory, force_upload, upload_mode, continue_partial):
    """
    start upload on a single run directory
    :param run_directory:
    :param force_upload:
    :param upload_mode
    :param continue_partial
    :return: exit code 0 or 1
    """
    return core.upload.upload_run_single_entry(run_directory, force_upload, upload_mode, continue_partial).exit_code


def upload_batch(batch_directory, force_upload, upload_mode, continue_partial):
    """
    Start uploading runs in the batch directory
    :param batch_directory:
    :param force_upload:
    :param upload_mode:
    :param continue_partial:
    :return: exit code 0 or 1
    """
    return core.upload.batch_upload_single_entry(batch_directory, force_upload, upload_mode, continue_partial).exit_code


# This is called when the program is run for the first time
if __name__ == "__main__":
    main()



import logging
# PyQt needs to be imported like this because for whatever reason they decided not to include a __all__ = [...]
import PyQt5.QtCore as QtCore

from iridauploader.core import api_handler


class QtHandler(logging.Handler):
    """
    Custom logging handler
    Overrides the default emit function to call the message writers emit instead

    A non Qt object cannot execute Qt emits so we wrap this handler around the MessageWriter
    """
    def __init__(self):
        logging.Handler.__init__(self)
        self._message_writer = MessageWriter()

    def emit(self, record):
        record = self.format(record)
        if record:
            self._message_writer.write(record)

    @property
    def message_writer(self):
        return self._message_writer


class MessageWriter(QtCore.QObject):
    """
    Wrapper the logging handler can use to emit a message,
    This object is needed to hold the messageWritten signal that needs to belong to Qt
    """
    messageWritten = QtCore.pyqtSignal(str)

    def flush(self):
        pass

    def write(self, msg):
        self.messageWritten.emit(msg)


def is_connected_to_irida():
    """
    Attempts connection to IRIDA, raises IridaConnectionError if unable
    """
    api_handler.initialize_api_from_config()



import logging
import os

import iridauploader.model as model
import iridauploader.progress as progress

from iridauploader.parsers import BaseParser
from iridauploader.parsers import exceptions
from iridauploader.parsers import common
from iridauploader.parsers.miniseq import sample_parser, validation


class Parser(BaseParser):

    SAMPLE_SHEET_FILE_NAME = 'SampleSheet.csv'
    UPLOAD_COMPLETE_FILE_NAME = 'CompletedJobInfo.xml'

    def __init__(self, parser_type_name='miniseq'):
        """
        Initialize the Parser
        :param parser_type_name: string to be included in metadata of sequencing run for type identification in IRIDA
        """
        super().__init__(
            parser_type_name=parser_type_name,
            required_file_list=[
                Parser.SAMPLE_SHEET_FILE_NAME,
                Parser.UPLOAD_COMPLETE_FILE_NAME
            ])

    @staticmethod
    def get_relative_data_directory():
        """
        Returns path to the sequence file directory, relative to the Sample Sheet

        This is not used in the application but is useful for scripting and cloud deployment

        This includes a '*' character to be interpreted as a wildcard symbol,
        as this sequencer does not make consitant names for data directories, and the * must be gotten from the filesystem

        :return: a string which represents the concatenated path components, as per os.path.join
        """
        data_dir = os.path.join("Alignment_1", "*", "Fastq")
        return data_dir

    @staticmethod
    def get_full_data_directory(sample_sheet):
        """
        Returns the path to where the sequence data files can be found, including the sample_sheet directory

        Note, this hits the os, and as such is not to be used with cloud solutions.
        For cloud solutions, use get_relative_data_directory() and solve the actual path for your cloud environment

        :param sample_sheet: Sample sheet acts as the starting point for the data directory
        :return: a string which represents the concatenated path components, as per os.path.join
        """
        sample_sheet_dir = os.path.dirname(sample_sheet)
        partial_data_dir = os.path.join(sample_sheet_dir, "Alignment_1")
        # Verify the partial path exits, path could not exist if there was a sequencing error
        # Also, if someone runs the miniseq parser on a miseq directory, this is the failure point
        if not os.path.exists(partial_data_dir):
            raise exceptions.DirectoryError(
                ("The uploader was unable to find the data directory, Verify that the run directory is "
                 "undamaged, and that it is a MiniSeq sequencing run."), partial_data_dir)

        # get the directories [1] get the first directory [0]
        data_dir = os.path.join(partial_data_dir, next(os.walk(partial_data_dir))[1][0], "Fastq")
        return data_dir

    def find_runs(self, directory):
        """
        find a list of run directories in the directory given

        :param directory:
        :return: list of DirectoryStatus objects
        """
        logging.info("looking for runs in {}".format(directory))

        runs = []
        directory_list = common.find_directory_list(directory)
        for d in directory_list:
            runs.append(progress.get_directory_status(d, self.get_required_file_list()))

        return runs

    def find_single_run(self, directory):
        """
        Find a run in the base directory given

        :param directory:
        :return: DirectoryStatus object
        """
        logging.info("looking for run in {}".format(directory))

        return progress.get_directory_status(directory, self.get_required_file_list())

    @staticmethod
    def get_sample_sheet(directory):
        """
        gets the sample sheet file path from a given run directory

        :param directory:
        :return:
        """
        logging.info("Looking for sample sheet in {}".format(directory))

        # Checks if we can access to the given directory, return empty and log a warning if we cannot.
        if common.cannot_read_directory(directory):
            logging.error(("The directory is not accessible, can not parse samples from this directory {}"
                           "".format(directory), directory))
            raise exceptions.DirectoryError("The directory is not accessible, "
                                            "can not parse samples from this directory {}".format(directory), directory)

        sample_sheet_file_name = Parser.SAMPLE_SHEET_FILE_NAME
        file_list = common.get_file_list(directory)  # Gets the list of files in the directory
        if sample_sheet_file_name not in file_list:
            logging.error("No sample sheet file in the MiniSeq format found")
            raise exceptions.DirectoryError("The directory {} has no sample sheet file in the MiniSeq format"
                                            " with the name {}"
                                            .format(directory, sample_sheet_file_name), directory)
        else:
            logging.debug("Sample sheet found")
            return os.path.join(directory, sample_sheet_file_name)

    def get_sequencing_run(self, sample_sheet, run_data_directory=None, run_data_directory_file_list=None):
        """
        Does local validation on the integrity of the run directory / sample sheet

        Throws a ValidationError with a validation result attached if it cannot make a sequencing run

        :param sample_sheet: Sample Sheet File
        :param run_data_directory: Optional: Directory (including run directory) to data files.
                                   Can be provided for bypassing os calls when developing on cloud systems
        :param run_data_directory_file_list: Optional: List of files in data directory.
                                             Can be provided for bypassing os calls when developing on cloud systems
        :return: SequencingRun
        """

        # get data directory and file list
        validation_result = model.ValidationResult()

        try:
            if run_data_directory is None:
                run_data_directory = Parser.get_full_data_directory(sample_sheet)
            if run_data_directory_file_list is None:
                run_data_directory_file_list = common.get_file_list(run_data_directory)
        except exceptions.DirectoryError as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while parsing files")
            raise exceptions.ValidationError("Errors occurred while parsing files", validation_result)

        # Try to get the sample sheet, validate that the sample sheet is valid
        validation_result = validation.validate_sample_sheet(sample_sheet)
        if not validation_result.is_valid():
            logging.error("Errors occurred while getting sample sheet")
            raise exceptions.ValidationError("Errors occurred while getting sample sheet", validation_result)

        # Try to parse the meta data from the sample sheet, throw validation error if errors occur
        validation_result = model.ValidationResult()
        try:
            run_metadata = sample_parser.parse_metadata(sample_sheet)
        except exceptions.SampleSheetError as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while parsing metadata")
            raise exceptions.ValidationError("Errors occurred while parsing metadata", validation_result)

        # Try to build sequencing run from sample sheet & meta data, raise validation error if errors occur
        try:
            bool_paired_files = run_metadata["layoutType"] == "PAIRED_END"
            sample_list = sample_parser.parse_sample_list(sample_sheet,
                                                          run_data_directory,
                                                          run_data_directory_file_list,
                                                          bool_paired_files)
            sequencing_run = common.build_sequencing_run_from_samples(sample_list,
                                                                      run_metadata,
                                                                      self.get_parser_type_name())
        except exceptions.SequenceFileError as error:
            validation_result.add_error(error)
            logging.error("Errors occurred while building sequence run from sample sheet")
            raise exceptions.ValidationError("Errors occurred while building sequence run from sample sheet",
                                             validation_result)

        return sequencing_run
